/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-conf-sock-1.c,v 1.5 2013/10/23 22:13:41 ca Exp $")

#include "sm/string.h"
#include "sm/net.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"
#include "sm/sockcnf.h"
#define SM_SOCKCNFDEF 1
#include "sm/sockcnfdef.h"
#include "sm/io.h"

/*
 *  Just a test program to play around with libconf (sm_conf_type_union).
 */
#define SOCKS_MAX	4
typedef struct
{
	sockspec_T		 s_sockets[SOCKS_MAX];
	unsigned int	 s_sockets_n;
} mts_T;

static sm_conf_definition_T
mts_defs[] =
{
	{ SM_CONF_DEF_MAGIC, "socket", sm_conf_type_union,
		offsetof(mts_T, s_sockets), sizeof(sockspec_T), NULL, 0,
		sock_spec_definitions, NULL, NULL, "sockets"
		SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

	{ SM_CONF_DEF_MAGIC, "n", sm_conf_type_array_n,
		offsetof(mts_T, s_sockets_n), sizeof(unsigned int), NULL, 0,
		sock_spec_definitions, NULL, NULL, "number of sockets (implicit)"
		SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

	CONF_SENTINEL
};

static sm_conf_definition_T
mts_def[] =
{

	{ SM_CONF_DEF_MAGIC, "listen_socket", sm_conf_type_array,
		0, SOCKS_MAX, NULL, SM_CONF_FLAG_MULTIPLE|SM_CONF_FLAG_FLAT,
		mts_defs, NULL, NULL,
		"sockets to listen() on"
		SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

	CONF_SENTINEL
};

static mts_T mts;

static void
print_ipv4(ipv4_T ipv4, const char *name)
{
	sm_ret_T ret;
	sm_str_P ipv4s;

	ipv4s = sm_str_new(NULL, 20, 32);
	if (ipv4s == NULL) {
		printf("out of memory\n");
		return;
	}
	ret = sm_inet_ipv4str(ipv4, ipv4s);
	if (ret != SM_SUCCESS) {
		printf("sm_inet_ipv4str=0x%x\n", ret);
		return;
	}
	printf("%s= %s\n", name, sm_str_getdata(ipv4s));
}

#if MTA_NETINET6
static void
print_ipv6(ipv6_P ipv6, const char *name)
{
	sm_ret_T ret;
	sm_str_P ipv6s;

	ipv6s = sm_str_new(NULL, 46, 64);
	if (ipv6s == NULL) {
		printf("out of memory\n");
		return;
	}
	ret = sm_inet_ipv6str(ipv6, ipv6s);
	if (ret != SM_SUCCESS) {
		printf("sm_inet_ipv6str=0x%x\n", ret);
		return;
	}
	printf("%s= %s\n", name, sm_str_getdata(ipv6s));
}
#endif

static void
print_sock(sockspec_T sckt)
{
	switch (sckt.sckspc_type) {
	  case SOCK_TYPE_UNIX:
		printf("path=\"%s\"\n", sckt.sock_unix.unixsckspc_path);
		break;
	  case SOCK_TYPE_INET:
		printf("IPv4\n");
		printf("port=%hd\n", sckt.sock_inet.inetsckspc_port);
		print_ipv4(sckt.sock_inet.inetsckspc_addr, "address");
		break;
#if MTA_NETINET6
	  case SOCK_TYPE_INET6:
		printf("IPv6\n");
		printf("port=%hd\n", sckt.sock_inet6.inet6sckspc_port);
		print_ipv6(&sckt.sock_inet6.inet6sckspc_addr, "address");
		break;
#endif
	  default:
		break;
	}
}

static void
print_mts(mts_T *st)
{
	uint i;

	for (i = 0; i < st->s_sockets_n; i++)
		print_sock(st->s_sockets[i]);
}

static int
process(char const *name, FILE *fp, int show)
{
	sm_conf_T *stream;
	int err;
	char buf[SM_CONF_ERROR_BUFFER_SIZE];
	char const *e = NULL;

	if (((stream = sm_conf_new(name ? name : "*stdin*"))) == NULL) {
		sm_io_fprintf(smioerr, "error -- sm_conf_new() returns NULL!\n");
		return 1;
	}
	if ((err = sm_conf_read_FILE(stream, name, fp)) != 0) {
		sm_io_fprintf(smioerr, "%s: %s\n",
			name ? name : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));

		while ((e = sm_conf_syntax_error(stream, e)) != NULL)
			sm_io_fprintf(smioerr, "%s\n", e);

		sm_conf_destroy(stream);
		return 2;
	}

	if ((err = sm_conf_scan(stream, mts_def, 0, &mts)) != 0) {
		sm_io_fprintf(smioerr, "(while scanning) %s: %s\n",
			name ? name : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));
		while ((e = sm_conf_syntax_error(stream, e)) != NULL)
			sm_io_fprintf(smioerr, "%s\n", e);
		sm_conf_destroy(stream);
		return 3;
	}
	print_mts(&mts);
	if (show) {
		err = sm_conf_prt_conf(mts_def, &mts, smioout);
		sm_io_flush(smioout);
	}

#if 0
	sm_conf_destroy(stream);
#endif /* 0 */

	return 0;
}

static void
showversion(void)
{
	sm_io_fprintf(smioerr, "compile time options:\n"
#if MTA_NETINET6
	"MTA_NETINET6\n"
#endif
	);
}

int
main(int ac, char **av)
{
	int ai, c, ret, done, show;

	done = 0;
	show = 0;
	while ((c = getopt(ac, av, "f:sV")) != -1) {
		switch (c) {
		  case 'f':
			ret = process(optarg, NULL, show);
			if (ret != 0)
				return ret;
			done = 1;
			break;
		  case 's':
			show = 1;
			break;
		  case 'V':
			showversion();
			exit(0);
			break;
		}
	}
	ac -= optind;
	av += optind;
	if (done) {
		print_mts(&mts);
		return 0;
	}

	if (ac == 0)
		return process("*stdin*", stdin, show);

	for (ai = 0; ai < ac; ai++) {
		ret = process(av[ai], NULL, show);
		if (ret != 0)
			return ret;
	}
	return 0;
}
