/*
 * Copyright (c) 2011 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: conf.c,v 1.12 2014/01/30 16:04:23 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/time.h"
#include "sm/heap.h"
#include "sm/mta.h"
#include "sm/sm-conf.h"
#include "sm/units.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#include "dkim.h"

#if MTA_USE_DKIM

#include "sm/mail_parse.h"
#include "sm/maps.h"
#include "sm/map.h"
#include "sm/mapclasses.h"

/*
**  DKIM_MAP_INIT -- Initialize DKIM signing map
**
**	Parameters:
**		dkim_ctx -- DKIM context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_map_init(dkim_ctx_P dkim_ctx)
{
	uint n, u;
	sm_ret_T ret;
	sm_str_T key, data;
	size_t datalen;
	dkim_cnf_P dkim_cnf;

	ret = SM_SUCCESS;
	datalen = sizeof(void *);
	dkim_cnf = dkim_ctx->dkc_cnf;
	n = dkim_cnf->dkimcnf_n;
	for (u = 0; u < n; u++) {
		const char *domain;
		size_t domainlen;
		dkim_s_P dkim_s;

		dkim_s = &(dkim_cnf->dkimcnf_s[u]);
		domain = dkim_s->dks_from;
		if (NULL == domain)
			continue;
		domainlen = strlen(domain);
		sm_str_assign(key, NULL, (uchar *)domain, domainlen, domainlen + 2);
		sm_str_assign(data, NULL, (void *)dkim_s, datalen, datalen + 2);
		ret = sm_map_add(dkim_ctx->dkc_dkim_map, &key, &data, SMMAP_AFL_UNIQUE);
#if 0
		sm_log_write(ss_ctx->ssc_lctx, SS_LCAT_SERVER, SS_LMOD_SERVER,
			SM_LOG_DEBUG, 10,
			"sev=DBG, func=dkim_map_init, sm_map_add=%m, key=%s"
			, ret, domain);
#endif
		if (sm_is_err(ret))
			break;
	}

	return ret;
}

/*
**  DKIM_MAP_NEW -- Create DKIM signing map
**
**	Parameters:
**		dkim_ctx -- DKIM context
**		pmaps_ctx -- (pointer to) map context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_map_new(dkim_ctx_P dkim_ctx, sm_maps_P *pmaps_ctx)
{
	uint n;
	sm_ret_T ret;
	sm_cstr_P mtype, mname;
	dkim_cnf_P dkim_cnf;

	ret = SM_SUCCESS;
	dkim_cnf = dkim_ctx->dkc_cnf;
	n = dkim_cnf->dkimcnf_n;
	if (0 == n)
		return SM_SUCCESS;

	mtype = mname = NULL;
	if (NULL == *pmaps_ctx) {
		ret = sm_maps_init(pmaps_ctx);
		if (NULL == *pmaps_ctx) {
			ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
			goto error;
		}
	}

#define SS_DKIM_TYPE "bhtable"
#define SS_DKIM_NAME "dkim"

	ret = sm_bht_class_create(*pmaps_ctx);
	if (sm_is_err(ret))
		goto error;

	mtype = sm_cstr_scpyn0((const uchar *)SS_DKIM_TYPE, strlen(SS_DKIM_TYPE));
	if (NULL == mtype) {
		ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
		goto error;
	}
	mname = sm_cstr_scpyn0((const uchar *)SS_DKIM_NAME, strlen(SS_DKIM_NAME));
	if (NULL == mname) {
		ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
		goto error;
	}

	dkim_ctx->dkc_dkim_map = NULL;
	ret = sm_map_open(*pmaps_ctx, mname, mtype, 0, NULL
			, SMAP_MODE_RDONLY, &dkim_ctx->dkc_dkim_map
			, SMPO_HASH_NELEM, n * 2 + 1, SMPO_ELEM_MAX, 0
			, SMPO_INIT_CBF, dkim_map_init, SMPO_INIT_CBC, dkim_ctx
			, SMPO_END);
	if (sm_is_err(ret)) {
#if 0
		sm_log_write(ss_ctx->ssc_lctx, SS_LCAT_SERVER, SS_LMOD_SERVER,
			SM_LOG_ERROR, 0,
			"sev=ERROR, func=dkim_map_new, sm_map_open=%m, map=%C, type=%C"
			, ret, mname, mtype);
#endif
		goto error;
	}
	SM_CSTR_FREE(mname);
	SM_CSTR_FREE(mtype);
	return ret;

  error:
	dkim_ctx->dkc_dkim_map = NULL;
	SM_CSTR_FREE(mname);
	SM_CSTR_FREE(mtype);
	return ret;
}

/*
**  DKIM_CONF_INIT -- DKIM setup
**
**	Parameters:
**		ss_ctx -- SMTP Server context
**		pmaps_ctx -- (pointer to) map context
**
**	Returns:
**		usual return code
*/

sm_ret_T
dkim_conf_init(dkim_ctx_P dkim_ctx, sm_maps_P *pmaps_ctx)
{
	uint u;
	sm_ret_T ret;
	dkim_cnf_P dkim_cnf;
	dkim_s_P dkim_s;

	ret = SM_NOTDONE;
	dkim_cnf = dkim_ctx->dkc_cnf;

	dkim_cnf->dkimcnf_no_sign = (NULL == dkim_cnf->dkimcnf_sign[0]);
	dkim_cnf->dkimcnf_no_skip = (NULL == dkim_cnf->dkimcnf_skip[0]);
	if (0 == dkim_cnf->dkimcnf_hdrtlen)
		dkim_cnf->dkimcnf_hdrtlen = SM_TOTAL_HDR_LEN;
	if (0 == dkim_cnf->dkimcnf_hdrvlen)
		dkim_cnf->dkimcnf_hdrvlen = SM_HDR_LEN_MAX;

	/* set default values??? see RFC 4871*/

	if (0 == dkim_cnf->dkimcnf_n)
		return ret;

	ret = SM_SUCCESS;
	rfc2822_init();

	/* find default selector and check that there is at most one */
	for (u = 0; u < dkim_cnf->dkimcnf_n; u++) {
		dkim_s = &(dkim_cnf->dkimcnf_s[u]);
		if (NULL == dkim_s->dks_from) {
			if (dkim_ctx->dkc_dk_s_dflt != NULL) {
#if 0
				sm_log_write(ss_ctx->ssc_lctx,
					SS_LCAT_SERVER, SS_LMOD_SERVER,
					SM_LOG_ERROR, 7,
					"sev=ERROR, func=ss_dkim_init, status=more than one section without from option");
				ss_ctx->ssc_use_dkim = false;
#endif
				ret = sm_error_perm(SM_EM_DKIM, SM_C_E_INVALID);
				goto error;
			}
			dkim_ctx->dkc_dk_s_dflt = dkim_s;
		}
	}

	/* do we need a map? */
	if (dkim_cnf->dkimcnf_n > 1 || NULL == dkim_ctx->dkc_dk_s_dflt) {
		ret = dkim_map_new(dkim_ctx, pmaps_ctx);
		if (sm_is_err(ret))
			goto error;
	}

	/* load keys for each selector */
	for (u = 0; u < dkim_cnf->dkimcnf_n; u++) {
		dkim_s = &(dkim_cnf->dkimcnf_s[u]);
		if (dkim_s->dks_keyfile != NULL &&
			dkim_s->dks_selector != NULL &&
			dkim_s->dks_domain != NULL)
		{
			ret = dkim_readkey(dkim_s);
			if (sm_is_err(ret)) {
#if 0
				sm_log_write(ss_ctx->ssc_lctx,
					SS_LCAT_SERVER, SS_LMOD_SERVER,
					SM_LOG_ERROR, 7,
					"sev=ERROR, func=ss_dkim_init, key=%s, dkim_readkey=%m"
					, dkim_s->dks_keyfile, ret);
				ss_ctx->ssc_use_dkim = false;
#endif
				ret = ret;
			}
		}
	}
	return ret;

  error:
	return ret;
}

#endif /* MTA_USE_DKIM */
