/*
 * Copyright (c) 2010 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-rcpt-set-domain-0.c,v 1.6 2011/01/30 21:48:59 ca Exp $")
#include "sm/sysexits.h"
#include "sm/str.h"
#include "sm/io.h"
#include "sm/mta.h"
#include "sm/rfc2821.h"
#include "sm/rpool.h"

#define USE_RPOOL_NONE	0
#define USE_RPOOL_ONCE	1
#define USE_RPOOL_EACH	2

static int Verbose = 0;

static sm_ret_T
rcpt_set_domain(sm_str_P rcpt, sm_str_P r_domain, sm_rpool_P rpool)
{
	sm_ret_T ret;
	sm_a2821_T a_rcpt, a_domain;
	sm_a2821_P prcpt, pdomain;

	SM_REQUIRE(rcpt != NULL);
	ret = SM_SUCCESS;
	pdomain = &a_domain;	/* pdomain is just pointer to a_domain */
	prcpt = &a_rcpt;	/* prcpt is just pointer to a_rcpt */
	r_domain = NULL;
	A2821_INIT_RP(prcpt, rpool);
	A2821_INIT_RP(pdomain, rpool);
	ret = t2821_scan((sm_rdstr_P) rcpt, prcpt, 0);
	if (sm_is_err(ret))
		goto error;
	ret = t2821_parse(prcpt, R2821_CORRECT & ~R2821_AT);
	if (sm_is_err(ret))
		goto error;

	r_domain = sm_str_new(rpool, SM_MIN(MAXDOMAINLEN, sm_str_getlen(rcpt)),
					MAXDOMAINLEN + 2);
	if (r_domain == NULL) {
		ret = sm_error_temp(SM_EM_AQ, ENOMEM);
		goto error;
	}
	ret = t2821_extract_domain(rpool, prcpt, &pdomain);
	if (!sm_is_err(ret))
		ret = t2821_str(pdomain, r_domain, 0);
	if (sm_is_err(ret))
		goto error;
	sm_str2lower(r_domain);
	a2821_free(pdomain);
	a2821_free(prcpt);
	return ret;

  error:
	a2821_free(pdomain);
	a2821_free(prcpt);
	SM_STR_FREE(r_domain);
	return ret;
}

static sm_ret_T
rcpt_get_domain(sm_str_P rcpt, sm_str_P r_domain, sm_rpool_P rpool)
{
	size_t len;
	uint i;
	sm_ret_T ret;

	SM_REQUIRE(rcpt != NULL);
	ret = SM_SUCCESS;
	len = sm_str_getlen(rcpt);
	r_domain = sm_str_new(rpool, SM_MIN(MAXDOMAINLEN, len), MAXDOMAINLEN + 2);
	if (r_domain == NULL) {
		ret = sm_error_temp(SM_EM_AQ, ENOMEM);
		goto error;
	}

	/* ret = sm_str_lchr(rcpt, (const uchar *)"@", &i); */
	ret = sm_str_l1chr(rcpt, '@', &i);
	if (sm_is_err(ret))
		goto error;
	/* after "@" and before ">" */
	ret = sm_str_catpart(r_domain, (const sm_rdstr_P) rcpt, i + 1, len - 2);
	if (Verbose > 2)
		sm_io_fprintf(smioerr, "%S -> %S\n", rcpt, r_domain);
	sm_str2lower(r_domain);
	return ret;

  error:
	SM_STR_FREE(r_domain);
	return ret;
}


static sm_ret_T
rsd_perf(int r, const char *rcpt, int use_rpool, bool simple)
{
	sm_ret_T ret;
	size_t len;
	sm_str_P srcpt;
	sm_str_P r_domain;
	sm_rpool_P rpool;

	SM_REQUIRE(rcpt != NULL);
	len = strlen(rcpt);
	srcpt = sm_str_crt(NULL, (uchar *)rcpt, len, len + 2);
	SM_ASSERT(srcpt != NULL);
	ret = SM_SUCCESS;
	r_domain = NULL;

	rpool = NULL;
	if (USE_RPOOL_ONCE == use_rpool) {
		rpool = sm_rpool_new(NULL);
		SM_ASSERT(rpool != NULL);
	}
	while (r-- > 0) {
		if (USE_RPOOL_EACH == use_rpool) {
			rpool = sm_rpool_new(NULL);
			SM_ASSERT(rpool != NULL);
		}
		if (simple)
			ret = rcpt_get_domain(srcpt, r_domain, rpool);
		else
			ret = rcpt_set_domain(srcpt, r_domain, rpool);
		if (USE_RPOOL_EACH == use_rpool) {
			sm_rpool_delete(rpool);
			rpool = NULL;
		}
	}
	if (USE_RPOOL_ONCE == use_rpool)
		sm_rpool_delete(rpool);
	return ret;
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n", prg);
	sm_io_fprintf(smioerr, "Test rcpt parsing\n");
	sm_io_fprintf(smioerr, "options:\n");
	sm_io_fprintf(smioerr, "-L    use new rpool each iteration\n");
	sm_io_fprintf(smioerr, "-R    use one rpool for all iterations\n");
	sm_io_fprintf(smioerr, "-s    only get the domain (default: set)\n");
	sm_io_fprintf(smioerr, "-r n  repeat n times\n");
	exit(EX_USAGE);
}

int
main(int argc, char **argv)
{
	int c, l;
	sm_ret_T r;
	bool simple;
	int use_rpool;
	char *prg, *rcpt;

	prg = argv[0];
	l = 1000;
	use_rpool = USE_RPOOL_NONE;
	rcpt = "<user@example.com>";
	simple = false;
	while ((c = getopt(argc, argv, "a:LRr:sV")) != -1) {
		switch (c) {
		  case 'a':
			rcpt = strdup(optarg);
			break;
		  case 'L':
			use_rpool = USE_RPOOL_EACH;
			break;
		  case 'R':
			use_rpool = USE_RPOOL_ONCE;
			break;
		  case 'r':
			l = (int) strtoul(optarg, NULL, 0);
			break;
		  case 's':
			simple = true;
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(prg);
			return EX_USAGE;
		}
	}

/*
	sm_test_begin(argc, argv, "test recipient parsing perf 0");
*/
	r = rsd_perf(l, rcpt, use_rpool, simple);
/*
	return sm_test_end();
*/
	return 0;
}
